/*
 * Copyright 2016 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.dua.tests.
 * 
 * de.bsvrz.dua.tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dua.tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.dua.tests.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.dua.tests;

import com.google.common.base.Joiner;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Sets;
import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.sys.funclib.bitctrl.dua.ufd.modell.DUAUmfeldDatenSensor;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.operatingMessage.OperatingMessageInterface;
import de.bsvrz.sys.funclib.operatingMessage.OperatingMessageSink;
import de.kappich.pat.testumg.util.DavTestEnvironment;
import de.kappich.pat.testumg.util.DavTestUtil;
import de.kappich.pat.testumg.util.FakeParamApp;
import de.kappich.sys.funclib.csv.CsvReader;
import de.kappich.sys.funclib.csv.IterableCsvData;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.nio.charset.Charset;
import java.time.Instant;
import java.time.LocalTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * TBD Dokumentation
 *
 * @author Kappich Systemberatung
 */
public abstract class DuATestBase {
	public static final int FORMAT_LEN = 20;
	public static DavTestEnvironment testEnvironment;
	protected static FakeParamApp fakeParamApp;
	protected static ClientDavInterface _connection;
	public static DataModel _dataModel;
	private static final AtomicLong _timestamp = new AtomicLong(60000);
	private static Deque<OperatingMessageInterface> _operatingMessages = new ArrayDeque<>();

	static {
		OperatingMessageSink.register(new OperatingMessageSink() {
			@Override
			public void publish(final OperatingMessageInterface message) {
				_operatingMessages.addFirst(message);
			}
		});
	}

	private boolean header_printed = false;

	public static void dataEquals(@Nullable final Data d1, @Nullable final Data d2, ArrayDeque<String> stack,
			Set<String> errors, final Collection<String> toIgnore) {
		if (d1 == null && d2 == null)
			return;
		if (d1 == null) {
			if (toIgnore.contains(d2.getName()))
				return;
			// errors.add(Joiner.on("→").join(stack) + "\nSoll: <Nicht
			// vorhanden>\nIst: " + d2 +"\n");
			return;
		} else if (d2 == null) {
			if (toIgnore.contains(d1.getName()))
				return;
			errors.add(Joiner.on("→").join(stack) + "\nSoll: " + d1 + "\nIst:  <Nicht vorhanden>\n");
			return;
		}
		if (toIgnore.contains(d1.getName()))
			return;
		stack.addLast(d1.getName());
		try {
			if (d1.isPlain()) {
				String expected = d1.asTextValue().getText();
				String actual = d2.asTextValue().getText();
				if (d1.getName().equals("Index")) {
					expected = String.valueOf(Math.round(d1.asUnscaledValue().doubleValue() / 100.0) + "%");
					actual = String.valueOf(Math.round(d2.asUnscaledValue().doubleValue() / 100.0) + "%");
				}
				if (!expected.equals(actual)) {
					errors.add(Joiner.on("→").join(stack) + "\nSoll: " + expected + "\nIst:  " + actual + "\n");
				}
				return;
			}
			Iterator<Data> d1It = d1.iterator();
			Iterator<Data> d2It = d2.iterator();
			while (d1It.hasNext() || d2It.hasNext()) {
				final Data data = d1It.next();
				final Data data1 = d2It.next();
				dataEquals(data, data1, stack, errors, toIgnore);
			}
		} finally {
			stack.removeLast();
		}
	}

	@Before
	public void setUp() throws Exception {
		testEnvironment = new DavTestEnvironment("ERROR");
		fakeParamApp = testEnvironment.createFakeParamApp();
		testEnvironment.startDav(getConfigurationAreas());
		_connection = testEnvironment.startDavConnection(null);
		_dataModel = _connection.getDataModel();

		// Workaround für statisches Verhalten der Funclib BitCtrl
		Method resetCache = DUAUmfeldDatenSensor.class.getDeclaredMethod("resetCache");
		resetCache.setAccessible(true);
		resetCache.invoke(null);
	}

	protected abstract String[] getConfigurationAreas();

	@After
	public void tearDown() throws Exception {
		_connection.disconnect(false, "");
		testEnvironment.stopAll(true);

	}

	public static OperatingMessageInterface pollMessage() {
		return _operatingMessages.pollLast();
	}

	protected void startTestCase(final String testCase, final SystemObject[] testObjectsIn,
			final SystemObject[] testObjectsOut, final DataDescription ddIn, final DataDescription ddOut,
			final ColumnLayout columnLayout) throws Exception {
		startTestCase(testCase, testObjectsIn, testObjectsOut, new ComplexDataDescription(ddIn),
				new ComplexDataDescription(ddOut), columnLayout);
	}

	protected void startTestCase(final String testCase, final SystemObject[] testObjectsIn,
			final SystemObject[] testObjectsOut, final DataDescription ddIn, final ComplexDataDescription ddOut,
			final ColumnLayout columnLayout) throws Exception {
		startTestCase(testCase, testObjectsIn, testObjectsOut, new ComplexDataDescription(ddIn),
				new ComplexDataDescription(ddOut), columnLayout);
	}

	protected void startTestCase(final String testCase, final SystemObject[] testObjectsIn,
			final SystemObject[] testObjectsOut, final ComplexDataDescription ddIn, final DataDescription ddOut,
			final ColumnLayout columnLayout) throws Exception {
		startTestCase(testCase, testObjectsIn, testObjectsOut, new ComplexDataDescription(ddIn),
				new ComplexDataDescription(ddOut), columnLayout);
	}

	protected void startTestCase(final String testCase, final SystemObject[] testObjectsIn,
			final SystemObject[] testObjectsOut, final ComplexDataDescription ddIn, final ComplexDataDescription ddOut,
			final ColumnLayout columnLayout) throws Exception {
		startTestCase(testCase, listify(testObjectsIn, ddIn), listify(testObjectsOut, ddOut), columnLayout);
	}

	private List<DuADataIdentification> listify(final SystemObject[] objects, ComplexDataDescription dataDescription) {
		return ImmutableList.of(new DuADataIdentification(objects, dataDescription));
	}

	protected void startTestCase(final String testCase, final List<DuADataIdentification> testObjectsIn,
			final List<DuADataIdentification> testObjectsOut, final ColumnLayout columnLayout) throws Exception {
		_operatingMessages.clear();
		Thread.sleep(1000); // Starten und initialisieren der Applikation
							// ermöglichen
		TestData testData = readTestData(testCase, columnLayout.groupingEnabled());
		List<List<List<String>>> dataSets = testData.getDataSets();
		int inputStartCol = testData.getInputStartCol();
		int outputStartCol = testData.getOutputStartCol();

		int numErrors = 0;

		_timestamp.set(columnLayout.getIntervallLength() * (1 + columnLayout.getOutOffset()));

		final Map<Long, ExpectedAndActualData> results = new TreeMap<>();

		for (int i = 1; i < dataSets.size(); i++) {
			List<List<String>> header = dataSets.get(0);
			List<List<String>> values = dataSets.get(i);
			int outIndex = i - columnLayout.getOutOffset();
			List<List<String>> valuesOut;
			if (outIndex > 0) {
				valuesOut = dataSets.get(outIndex);
			} else {
				valuesOut = Collections.emptyList();
			}
			if (values.isEmpty())
				continue;
			long sendTime = _timestamp.get();

			printLine(values.get(0).get(0));

			long receiveTime = sendTime - columnLayout.getOutOffset() * columnLayout.getIntervallLength();
			List<ResultData> in = readData(testObjectsIn, header, values, inputStartCol, columnLayout, sendTime, true);
			List<ResultData> expected = readData(testObjectsOut, header, valuesOut, outputStartCol, columnLayout,
					receiveTime, false);
			List<String> messageExpected = readExpectedMessages(testObjectsOut, header, valuesOut, columnLayout,
					receiveTime);
			ResultData[] toSend = in.toArray(new ResultData[in.size()]);
			sendData(toSend);
			for (ResultData resultData : toSend) {
				printSent(resultData, Collections.emptySet());
			}
			for (ResultData resultData : expected) {
				printExpected(resultData, columnLayout.getIgnored());
			}
			for (DuADataIdentification duADataIdentification : testObjectsOut) {
				for (SystemObject object : duADataIdentification.getObjects()) {
					for (DataDescription subDD : duADataIdentification.getDataDescription()) {
						ResultData e = readData(object, subDD);
						while (e != null) {
							getResult(results, e.getDataTime()).getActual().add(e);
							printReceived(e, Collections.emptySet());
							e = readData(object, subDD);
						}
					}
				}
			}
			OperatingMessageInterface message = pollMessage();

			getResult(results, receiveTime).getExpected().addAll(expected);
			getResult(results, receiveTime).setMessage(message == null ? null : message.getMessage());
			getResult(results, receiveTime).getMessageExpected().addAll(messageExpected);
			getResult(results, receiveTime).setLineNumber(values.get(0).get(0));

			_timestamp.addAndGet(columnLayout.getIntervallLength());

		}

		System.out.println("\nListe der Fehler: ");
		for (ExpectedAndActualData result : results.values()) {
			Set<String> errors = new LinkedHashSet<>();
			String message = result.getMessage();
			for (String expectedMessage : result.getMessageExpected()) {
				if ((message != null) || (expectedMessage != null)) {
					if (expectedMessage == null) {
						errors.add("Unerwartete Betriebsmeldung: " + message);
					} else if (message == null) {
						errors.add("Fehlende Betriebsmeldung: " + expectedMessage);
					} else if (!message.contains(expectedMessage)) {
						errors.add("Falsche Betriebsmeldung. Soll:\n" + expectedMessage + "\nIst:" + message);
					} else {
						System.out.println(
								"Betriebsmeldung empfangen. Z: " + result.getLineNumber() + " Nachricht: " + message);
					}
				}
			}
			Map<DuADataIdentification, ResultData> expected = collect(result.getExpected());
			Map<DuADataIdentification, ResultData> actual = collect(result.getActual());

			Sets.SetView<DuADataIdentification> ids = Sets.union(expected.keySet(), actual.keySet());

			for (DuADataIdentification id : ids) {
				ResultData exData = expected.get(id);
				ResultData actData = actual.get(id);
				DuATestBase.dataEquals(exData == null ? null : exData.getData(),
						actData == null ? null : actData.getData(), new ArrayDeque<>(), errors,
						columnLayout.getIgnored());
			}

			if (!errors.isEmpty()) {
				if (!columnLayout.getIgnoredLines().contains(result.getLineNumber())) {
					System.out.println("Zeile: " + result.getLineNumber());
					for (String error : errors) {
						System.out.println(error);
						numErrors++;
					}
				}
			}
		}

		System.out.println("Anzahl Fehler: " + numErrors);

		Assert.assertEquals(0, numErrors);
	}

	private void printLine(final String lineNumber) {
		header_printed = false;
		print("", "");
		print("    Zeile: ", lineNumber);
	}

	private void printSent(final ResultData resultData, final Collection<String> toIgnore) {
		print(" Gesendet: ", dataToString(resultData, toIgnore));
	}

	private void printExpected(final ResultData resultData, final Collection<String> toIgnore) {
		print(" Erwartet: ", dataToString(resultData, toIgnore));
	}

	private void printReceived(final ResultData resultData, final Collection<String> toIgnore) {
		print("Empfangen: ", dataToString(resultData, toIgnore));
	}

	private void print(final String s, final String data) {
		System.out.println(s + data);
	}

	private String dataToString(final ResultData resultData, final Collection<String> toIgnore) {
		StringBuilder stringBuilder = new StringBuilder();
		LocalTime localTime = Instant.ofEpochMilli(resultData.getDataTime()).atZone(ZoneId.systemDefault())
				.toLocalTime();
		stringBuilder.append(DateTimeFormatter.ofLocalizedTime(FormatStyle.SHORT).format(localTime));
		appendFormatted(stringBuilder, resultData.getObject().getPidOrId());
		appendFormatted(stringBuilder, resultData.getDataDescription().getAttributeGroup().getPidOrId());
		appendFormatted(stringBuilder, resultData.getDataDescription().getAspect().getPidOrId());
		if (!resultData.hasData()) {
			stringBuilder.append("; <").append(resultData.getDataState()).append(">");
		} else {
			dataToString(resultData.getData(), stringBuilder, toIgnore, false);
		}
		return stringBuilder.toString();
	}

	private void dataToString(final Data data, final StringBuilder stringBuilder, final Collection<String> toIgnore,
			boolean ignored) {
		if (toIgnore.contains(data.getName())) {
			ignored = true;
		}
		if (data.isPlain()) {
			if (ignored) {
				appendFormatted(stringBuilder, "?");
			} else {
				appendFormatted(stringBuilder, data.asTextValue().getValueText());
			}
		} else {
			for (Data data1 : data) {
				dataToString(data1, stringBuilder, toIgnore, ignored);
			}
		}
	}

	private void appendFormatted(final StringBuilder stringBuilder, final String text) {
		int length = stringBuilder.length();
		stringBuilder.append("; ").append(String.format("%-" + FORMAT_LEN + "s", text));
		if (stringBuilder.length() > length + FORMAT_LEN + 2) {
			stringBuilder.setLength(length + FORMAT_LEN + 1);
			stringBuilder.append("…");
		}
	}

	private Map<DuADataIdentification, ResultData> collect(final List<ResultData> resultDatas) {
		try {
			return resultDatas.stream().collect(Collectors.toMap(new Function<ResultData, DuADataIdentification>() {
				@Override
				public DuADataIdentification apply(final ResultData t) {
					return new DuADataIdentification(t.getObject(), t.getDataDescription());
				}
			}, t -> t

			));
		} catch (Exception e) {
			System.out.println("e = " + e);
			return Collections.emptyMap();
		}
	}

	private ExpectedAndActualData getResult(final Map<Long, ExpectedAndActualData> results, final long receiveTime) {
		ExpectedAndActualData v = results.get(receiveTime);
		if (v != null)
			return v;
		ExpectedAndActualData value = new ExpectedAndActualData();
		results.put(receiveTime, value);
		return value;
	}

	private ResultData readData(final SystemObject object, final DataDescription subDD) throws InterruptedException {
		try {
			return DavTestUtil.readData(object, subDD, getDefaultTimeOut());
		} catch (AssertionError e) {
			return null;
		}
	}

	/**
	 * liefert den Standard-Timeoutwert beim Lesen von Datenverteilerdaten in
	 * Millisekunden.
	 * 
	 * @return der Timeout-Wert in Millisekunden
	 */
	protected long getDefaultTimeOut() {
		return 50;
	}

	public TestData readTestData(final String testCase, final boolean groupingEnabled) throws IOException {
		List<List<List<String>>> dataSets;
		int inputStartCol;
		int outputStartCol;
		try (InputStream inputStream = getClass().getResourceAsStream("testdata/" + testCase)) {
			CsvReader csvReader = new CsvReader(Charset.forName("UTF-8"), inputStream, ',', '"');
			IterableCsvData data = csvReader.read(null);
			inputStartCol = -1;
			outputStartCol = -1;
			dataSets = new ArrayList<>();
			final List<List<String>> group = new ArrayList<>();
			for (final IterableCsvData.CsvRow csvRow : data) {
				List<String> strings = csvRow.asList();
				if (strings.get(0).equals("Zeile")) {
					inputStartCol = strings.indexOf("Eingangsdaten");
					outputStartCol = Collections.max(ImmutableList.of(strings.indexOf("Soll"),
							strings.indexOf("Soll Formeln"), strings.indexOf("Soll erwartet")));
					continue;
				}
				if (inputStartCol != -1) {
					if (strings.get(0).isEmpty() && (dataSets.isEmpty() || groupingEnabled)) {
						group.add(ImmutableList.copyOf(strings));
					} else {
						if (!group.isEmpty()) {
							dataSets.add(ImmutableList.copyOf(group));
							group.clear();
						}
						group.add(ImmutableList.copyOf(strings));
					}
				}
			}
			dataSets.add(ImmutableList.copyOf(group));
		}
		return new TestData(dataSets, inputStartCol, outputStartCol);
	}

	public void sendData(final ResultData... resultDatas) throws SendSubscriptionNotConfirmed {
		_connection.sendData(resultDatas);
		try {
			Thread.sleep(200);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	private List<ResultData> readData(final List<DuADataIdentification> dataIdentifications,
			final List<List<String>> header, final List<List<String>> values, final int startColumn,
			final ColumnLayout layout, final long timestamp, boolean in) {
		final List<ResultData> resultDatas = new ArrayList<ResultData>();
		for (DuADataIdentification dataIdentification : dataIdentifications) {
			SystemObject[] testObjects = dataIdentification.getObjects();
			for (int i = 0; i < values.size() && i < testObjects.length; i++) {
				List<String> fullRow = values.get(i);
				if (fullRow.stream().filter(s -> s != null && !s.isEmpty()).count() < 3) {
					ComplexDataDescription dataDescriptions = dataIdentification.getDataDescription();
					for (DataDescription description : dataDescriptions) {
						resultDatas.add(new ResultData(testObjects[i], description, timestamp, null));
					}
				} else {
					final List<String> row = fullRow.subList(startColumn, fullRow.size());
					if (row.stream().filter(s -> s != null && !s.isEmpty()).count() == 0) {
						ComplexDataDescription dataDescriptions = dataIdentification.getDataDescription();
						for (DataDescription description : dataDescriptions) {
							resultDatas.add(new ResultData(testObjects[i], description, timestamp, null));
						}
					} else {
						ComplexDataDescription dataDescriptions = dataIdentification.getDataDescription();
						for (DataDescription description : dataDescriptions) {
							Data data = _connection.createData(description.getAttributeGroup());
							resetData(data);
							int offset = 0;
							if (header.get(0).get(startColumn).equals("T")) {
								try {
									String unit = header.get(header.size() - 1).get(startColumn).substring(0, 1)
											.toLowerCase();
									switch (unit) {
									case "s":
										data.getTimeValue("T").setMillis(
												(long) (Double.parseDouble(row.get(0).replaceAll(",", ".")) * 1000));
										break;
									case "m":
										data.getTimeValue("T").setMillis(
												(long) (Double.parseDouble(row.get(0).replaceAll(",", ".")) * 60000));
										break;

									}
								} catch (NoSuchElementException e) {
								}
								offset = 1;
							}
							int n = 0;
							while (true) {
								int realCol = offset + n * layout.getColumnCount(in);
								if (startColumn + realCol >= header.get(0).size())
									break;
								String type = header.get(Math.max(0, header.size() - 2)).get(startColumn + realCol);
								if (type.isEmpty())
									break;
								List<Data> datas = null;
								if (type.equals("Wert")) {
									type = description.getAttributeGroup().getPid().substring(8);
								}
								datas = layout.getSubDatas(type, data);
								for (Data data1 : datas) {
									layout.setValues(testObjects[i], data1, row, realCol, type, in);
								}
								n++;
							}
							if (!data.isDefined() && in) {
								System.err.println("Ignoriere undefiniertes Datum: " + data);
							} else {
								resultDatas.add(new ResultData(testObjects[i], description, timestamp, data));
							}
						}
					}
				}
			}
		}

		return resultDatas;
	}

	private List<String> readExpectedMessages(final List<DuADataIdentification> duADataIdentifications,
			final List<List<String>> header, final List<List<String>> values, final ColumnLayout layout,
			final long timestamp) {
		final List<String> results = new ArrayList<String>();
		for (DuADataIdentification duADataIdentification : duADataIdentifications) {
			for (DataDescription dataDescription : duADataIdentification.getDataDescription()) {
				for (int i = 0; i < values.size() && i < duADataIdentification.getObjects().length; i++) {
					List<String> fullRow = values.get(i);
					if (fullRow.stream().filter(s -> s != null && !s.isEmpty()).count() < 3) {
						results.add(null);
					} else {
						String expectedMessageText = layout.getExpectedMessageText(fullRow, dataDescription, header,
								timestamp);
						if (expectedMessageText != null && expectedMessageText.isEmpty()) {
							expectedMessageText = null;
						}
						results.add(expectedMessageText);
					}
				}
			}
		}

		return results;
	}

	public void resetData(final Data data) {
		if (data.isPlain()) {
			if (!data.isDefined()) {
				AttributeType attributeType = data.getAttributeType();
				if (attributeType instanceof IntegerAttributeType) {
					IntegerAttributeType integerAttributeType = (IntegerAttributeType) attributeType;
					List<IntegerValueState> states = integerAttributeType.getStates();
					if (!states.isEmpty()) {
						data.asUnscaledValue().set(states.get(states.size() - 1).getValue());
					} else {
						data.asUnscaledValue().set(integerAttributeType.getRange().getMinimum());
					}
				}
			}
		} else {
			for (Data inner : data) {
				resetData(inner);
			}
		}
	}

	public class TestData {
		private final List<List<List<String>>> _dataSets;
		private final int _inputStartCol;
		private final int _outputStartCol;

		public TestData(final List<List<List<String>>> dataSets, final int inputStartCol, final int outputStartCol) {
			_dataSets = dataSets;
			_inputStartCol = inputStartCol;
			_outputStartCol = outputStartCol;
		}

		public List<List<List<String>>> getDataSets() {
			return _dataSets;
		}

		public int getInputStartCol() {
			return _inputStartCol;
		}

		public int getOutputStartCol() {
			return _outputStartCol;
		}
	}

	private class ExpectedAndActualData {
		private String _lineNumber;
		private String _message;
		private final List<ResultData> _expected = new ArrayList<>();
		private final List<ResultData> _actual = new ArrayList<>();
		private final Set<String> _messageExpected = new LinkedHashSet<>();

		public ExpectedAndActualData() {
		}

		public List<ResultData> getExpected() {
			return _expected;
		}

		public List<ResultData> getActual() {
			return _actual;
		}

		public String getLineNumber() {
			return _lineNumber;
		}

		public String getMessage() {
			return _message;
		}

		public Set<String> getMessageExpected() {
			return _messageExpected;
		}

		public void setLineNumber(final String lineNumber) {
			_lineNumber = lineNumber;
		}

		public void setMessage(final String message) {
			_message = message;
		}
	}
}
